@@ -83,6 +83,8 @@ static char *audiobuf = NULL;
 static snd_pcm_uframes_t chunk_size = 0;
 static unsigned period_time = 0;
 static unsigned buffer_time = 0;
+static snd_pcm_uframes_t period_frames = 0;
+static snd_pcm_uframes_t buffer_frames = 0;
 static int avail_min = -1;
 static int start_delay = 0;
 static int stop_delay = 0;
@@ -93,7 +95,7 @@ static size_t chunk_bytes;
 static snd_output_t *log;
 
 static int fd = -1;
-static size_t pbrec_count = (size_t)-1, fdcount;
+static off64_t pbrec_count = (size_t)-1, fdcount;
 static int vocmajor, vocminor;
 
 /* needed prototypes */
@@ -142,8 +144,8 @@ static void usage(char *command)
 	fprintf(stderr,
 "Usage: %s [OPTION]... [FILE]...\n"
 "\n"
-"--help                  help\n"
-"--version               print current version\n"
+"-h, --help              help\n"
+"    --version           print current version\n"
 "-l, --list-devices      list all soundcards and digital audio devices\n"
 "-L, --list-pcms         list all PCMs defined\n"
 "-D, --device=NAME       select PCM by name\n"
@@ -158,6 +160,8 @@ static void usage(char *command)
 "-N, --nonblock          nonblocking mode\n"
 "-F, --period-time=#     distance between interrupts is # microseconds\n"
 "-B, --buffer-time=#     buffer duration is # microseconds\n"
+"    --period-size=#     distance between interrupts is # frames\n"
+"    --buffer-size=#     buffer duration is # frames\n"
 "-A, --avail-min=#       min available space for wakeup is # microseconds\n"
 "-R, --start-delay=#     delay for automatic PCM start is # microseconds \n"
 "                        (relative to buffer size if <= 0)\n"
@@ -265,7 +269,7 @@ static void pcm_list(void)
 
 static void version(void)
 {
-	fprintf(stderr, "%s: version " SND_UTIL_VERSION_STR " by Jaroslav Kysela <perex@suse.cz>", command);
+	fprintf(stderr, "%s: version " SND_UTIL_VERSION_STR " by Jaroslav Kysela <perex@suse.cz>\n", command);
 }
 
 static void signal_handler(int sig)
@@ -287,15 +291,18 @@ static void signal_handler(int sig)
 	exit(EXIT_FAILURE);
 }
 
-#define OPT_HELP 1
-#define OPT_VERSION 2
+enum {
+	OPT_VERSION = 1,
+	OPT_PERIOD_SIZE,
+	OPT_BUFFER_SIZE
+};
 
 int main(int argc, char *argv[])
 {
 	int option_index;
-	char *short_options = "lLD:qt:c:f:r:d:s:MNF:A:X:R:T:B:vIPC";
+	char *short_options = "hlLD:qt:c:f:r:d:s:MNF:A:X:R:T:B:vIPC";
 	static struct option long_options[] = {
-		{"help", 0, 0, OPT_HELP},
+		{"help", 0, 0, 'h'},
 		{"version", 0, 0, OPT_VERSION},
 		{"list-devices", 0, 0, 'l'},
 		{"list-pcms", 0, 0, 'L'},
@@ -310,10 +317,12 @@ int main(int argc, char *argv[])
 		{"mmap", 0, 0, 'M'},
 		{"nonblock", 0, 0, 'N'},
 		{"period-time", 1, 0, 'F'},
+		{"period-size", 1, 0, OPT_PERIOD_SIZE},
 		{"avail-min", 1, 0, 'A'},
 		{"start-delay", 1, 0, 'R'},
 		{"stop-delay", 1, 0, 'T'},
 		{"buffer-time", 1, 0, 'B'},
+		{"buffer-size", 1, 0, OPT_BUFFER_SIZE},
 		{"verbose", 0, 0, 'v'},
 		{"separate-channels", 0, 0, 'I'},
 		{"playback", 0, 0, 'P'},
@@ -325,6 +334,8 @@ int main(int argc, char *argv[])
 	int do_device_list = 0, do_pcm_list = 0;
 	snd_pcm_info_t *info;
 
+	snd_pcm_info_alloca(&info);
+
 	err = snd_output_stdio_attach(&log, stderr, 0);
 	assert(err >= 0);
 
@@ -350,7 +361,7 @@ int main(int argc, char *argv[])
 
 	while ((c = getopt_long(argc, argv, short_options, long_options, &option_index)) != -1) {
 		switch (c) {
-		case OPT_HELP:
+		case 'h':
 			usage(command);
 			return 0;
 		case OPT_VERSION:
@@ -414,7 +425,7 @@ int main(int argc, char *argv[])
 			if (tmp < 300)
 				tmp *= 1000;
 			rhwparams.rate = tmp;
-			if (tmp < 2000 || tmp > 128000) {
+			if (tmp < 2000 || tmp > 192000) {
 				error("bad speed value %i", tmp);
 				return 1;
 			}
@@ -435,6 +446,12 @@ int main(int argc, char *argv[])
 		case 'B':
 			buffer_time = atoi(optarg);
 			break;
+		case OPT_PERIOD_SIZE:
+			period_frames = atoi(optarg);
+			break;
+		case OPT_BUFFER_SIZE:
+			buffer_frames = atoi(optarg);
+			break;
 		case 'A':
 			avail_min = atoi(optarg);
 			break;
@@ -473,9 +490,11 @@ int main(int argc, char *argv[])
 	if (do_device_list) {
 		if (do_pcm_list) pcm_list();
 		device_list();
+		snd_config_update_free_global();
 		return 0;
 	} else if (do_pcm_list) {
 		pcm_list();
+		snd_config_update_free_global();
 		return 0;
 	}
 
@@ -485,7 +504,6 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
-	snd_pcm_info_alloca(&info);
 	if ((err = snd_pcm_info(handle, info)) < 0) {
 		error("info error: %s", snd_strerror(err));
 		return 1;
@@ -547,6 +565,7 @@ int main(int argc, char *argv[])
 	snd_pcm_close(handle);
 	free(audiobuf);
 	snd_output_close(log);
+	snd_config_update_free_global();
 	return EXIT_SUCCESS;
 }
 
@@ -829,15 +848,28 @@ static void set_params(void)
 		}
 	}
 	rate = hwparams.rate;
-	if (buffer_time == 0)
+	if (buffer_time == 0 && buffer_frames == 0)
 		buffer_time = 500000;
-	err = snd_pcm_hw_params_set_buffer_time_near(handle, params,
-						     &buffer_time, 0);
+	if (period_time == 0 && period_frames == 0) {
+		if (buffer_time > 0)
+			period_time = buffer_time / 4;
+		else
+			period_frames = buffer_frames / 4;
+	}
+	if (period_time > 0)
+		err = snd_pcm_hw_params_set_period_time_near(handle, params,
+							     &period_time, 0);
+	else
+		err = snd_pcm_hw_params_set_period_size_near(handle, params,
+							     &period_frames, 0);
 	assert(err >= 0);
-	if (period_time == 0)
-		period_time = buffer_time / 4;
-	err = snd_pcm_hw_params_set_period_time_near(handle, params,
-						     &period_time, 0);
+	if (buffer_time > 0) {
+		err = snd_pcm_hw_params_set_buffer_time_near(handle, params,
+							     &buffer_time, 0);
+	} else {
+		err = snd_pcm_hw_params_set_buffer_size_near(handle, params,
+							     &buffer_frames);
+	}
 	assert(err >= 0);
 	err = snd_pcm_hw_params(handle, params);
 	if (err < 0) {
@@ -937,7 +969,9 @@ static void xrun(void)
 		gettimeofday(&now, 0);
 		snd_pcm_status_get_trigger_tstamp(status, &tstamp);
 		timersub(&now, &tstamp, &diff);
-		fprintf(stderr, "xrun!!! (at least %.3f ms long)\n", diff.tv_sec * 1000 + diff.tv_usec / 1000.0);
+		fprintf(stderr, "%s!!! (at least %.3f ms long)\n",
+			stream == SND_PCM_STREAM_PLAYBACK ? "underrun" : "overrun",
+			diff.tv_sec * 1000 + diff.tv_usec / 1000.0);
 		if (verbose) {
 			fprintf(stderr, "Status:\n");
 			snd_pcm_status_dump(status, log);
@@ -947,6 +981,23 @@ static void xrun(void)
 			exit(EXIT_FAILURE);
 		}
 		return;		/* ok, data should be accepted again */
+	} if (snd_pcm_status_get_state(status) == SND_PCM_STATE_DRAINING) {
+		if (verbose) {
+			fprintf(stderr, "Status(DRAINING):\n");
+			snd_pcm_status_dump(status, log);
+		}
+		if (stream == SND_PCM_STREAM_CAPTURE) {
+			fprintf(stderr, "capture stream format change? attempting recover...\n");
+			if ((res = snd_pcm_prepare(handle))<0) {
+				error("xrun(DRAINING): prepare error: %s", snd_strerror(res));
+				exit(EXIT_FAILURE);
+			}
+			return;
+		}
+	}
+	if (verbose) {
+		fprintf(stderr, "Status(R/W):\n");
+		snd_pcm_status_dump(status, log);
 	}
 	error("read/write error, state = %s", snd_pcm_state_name(snd_pcm_status_get_state(status)));
 	exit(EXIT_FAILURE);
@@ -987,6 +1038,7 @@ static void compute_max_peak(u_char *data, size_t count)
 		default: val = 0; step = 1; break;
 		}
 		data += step;
+		val = abs(val);
 		if (max_peak < val)
 			max_peak = val;
 	}
@@ -994,7 +1046,10 @@ static void compute_max_peak(u_char *data, size_t count)
 	if (max <= 0)
 		max = 0x7fffffff;
 	printf("Max peak (%li samples): %05i (0x%04x) ", (long)ocount, max_peak, max_peak);
-	perc = max_peak / (max / 100);
+	if (bits_per_sample > 16)
+		perc = max_peak / (max / 100);
+	else
+		perc = max_peak * 100 / max;
 	for (val = 0; val < 20; val++)
 		if (val <= perc / 5)
 			putc('#', stdout);
@@ -1238,7 +1293,7 @@ static void voc_play(int fd, int ofs, char *name)
 	char was_extended = 0, output = 0;
 	u_short *sp, repeat = 0;
 	size_t silence;
-	int filepos = 0;
+	off64_t filepos = 0;
 
 #define COUNT(x)	nextblock -= x; in_buffer -= x; data += x
 #define COUNT1(x)	in_buffer -= x; data += x
@@ -1369,7 +1424,7 @@ static void voc_play(int fd, int ofs, char *name)
 				d_printf("Repeat loop %d times\n", repeat);
 #endif
 				if (filepos >= 0) {	/* if < 0, one seek fails, why test another */
-					if ((filepos = lseek(fd, 0, 1)) < 0) {
+					if ((filepos = lseek64(fd, 0, 1)) < 0) {
 						error("can't play loops; %s isn't seekable\n", name);
 						repeat = 0;
 					} else {
@@ -1391,7 +1446,7 @@ static void voc_play(int fd, int ofs, char *name)
 					else
 						d_printf("Neverending loop\n");
 #endif
-					lseek(fd, filepos, 0);
+					lseek64(fd, filepos, 0);
 					in_buffer = 0;	/* clear the buffer */
 					goto Fill_the_buffer;
 				}
@@ -1455,16 +1510,15 @@ static void init_raw_data(void)
 }
 
 /* calculate the data count to read from/to dsp */
-static size_t calc_count(void)
+static off64_t calc_count(void)
 {
-	size_t count;
+	off64_t count;
 
 	if (!timelimit) {
-		count = (size_t)-1;
+		count = (off64_t)-1;
 	} else {
-		count = snd_pcm_format_size(hwparams.format,
-					    timelimit * hwparams.rate *
-					    hwparams.channels);
+		count = snd_pcm_format_size(hwparams.format, hwparams.rate * hwparams.channels);
+		count *= (off64_t)timelimit;
 	}
 	return count < pbrec_count ? count : pbrec_count;
 }
@@ -1622,7 +1676,7 @@ static void begin_au(int fd, size_t cnt)
 /* closing .VOC */
 static void end_voc(int fd)
 {
-	off_t length_seek;
+	off64_t length_seek;
 	VocBlockType bt;
 	size_t cnt;
 	char dummy = 0;		/* Write a Terminator */
@@ -1642,7 +1696,7 @@ static void end_voc(int fd)
 	bt.datalen = (u_char) (cnt & 0xFF);
 	bt.datalen_m = (u_char) ((cnt & 0xFF00) >> 8);
 	bt.datalen_h = (u_char) ((cnt & 0xFF0000) >> 16);
-	if (lseek(fd, length_seek, SEEK_SET) == length_seek)
+	if (lseek64(fd, length_seek, SEEK_SET) == length_seek)
 		write(fd, &bt, sizeof(VocBlockType));
 	if (fd != 1)
 		close(fd);
@@ -1657,7 +1711,7 @@ static void end_raw(int fd)
 static void end_wave(int fd)
 {				/* only close output */
 	WaveChunkHeader cd;
-	off_t length_seek;
+	off64_t length_seek;
 	u_int rifflen;
 	
 	length_seek = sizeof(WaveHeader) +
@@ -1667,9 +1721,9 @@ static void end_wave(int fd)
 	cd.length = fdcount > 0x7fffffff ? 0x7fffffff : LE_INT(fdcount);
 	rifflen = fdcount + 2*sizeof(WaveChunkHeader) + sizeof(WaveFmtBody) + 4;
 	rifflen = rifflen > 0x7fffffff ? 0x7fffffff : LE_INT(rifflen);
-	if (lseek(fd, 4, SEEK_SET) == 4)
+	if (lseek64(fd, 4, SEEK_SET) == 4)
 		write(fd, &rifflen, 4);
-	if (lseek(fd, length_seek, SEEK_SET) == length_seek)
+	if (lseek64(fd, length_seek, SEEK_SET) == length_seek)
 		write(fd, &cd, sizeof(WaveChunkHeader));
 	if (fd != 1)
 		close(fd);
@@ -1678,11 +1732,11 @@ static void end_wave(int fd)
 static void end_au(int fd)
 {				/* only close output */
 	AuHeader ah;
-	off_t length_seek;
+	off64_t length_seek;
 	
 	length_seek = (char *)&ah.data_size - (char *)&ah;
 	ah.data_size = fdcount > 0xffffffff ? 0xffffffff : BE_INT(fdcount);
-	if (lseek(fd, length_seek, SEEK_SET) == length_seek)
+	if (lseek64(fd, length_seek, SEEK_SET) == length_seek)
 		write(fd, &ah.data_size, sizeof(ah.data_size));
 	if (fd != 1)
 		close(fd);
@@ -1803,7 +1857,7 @@ static void playback(char *name)
 		fd = fileno(stdin);
 		name = "stdin";
 	} else {
-		if ((fd = open(name, O_RDONLY, 0)) == -1) {
+		if ((fd = open64(name, O_RDONLY, 0)) == -1) {
 			perror(name);
 			exit(EXIT_FAILURE);
 		}
@@ -1854,7 +1908,7 @@ static void capture(char *name)
 		name = "stdout";
 	} else {
 		remove(name);
-		if ((fd = open(name, O_WRONLY | O_CREAT, 0644)) == -1) {
+		if ((fd = open64(name, O_WRONLY | O_CREAT, 0644)) == -1) {
 			perror(name);
 			exit(EXIT_FAILURE);
 		}
